@@ -7,7 +7,6 @@ import ccxt.pro as ccxt
 from contextlib import suppress
 from ccxt import DECIMAL_PLACES
 
-# import ccxt.async_support as ccxt
 from decimal import Decimal, ROUND_DOWN
 from datetime import datetime, timedelta, timezone
 from tortoise import connections
@@ -29,6 +28,8 @@ class ExchangeEnum(str, Enum):
     BYBIT = "BYBIT"
     WOO = "WOO"
     HYPERLIQUID = "HYPERLIQUID"
+    BINANCE = "BINANCE"
+    KRAKEN = "KRAKEN"
 
 
 class MarketTypeEnum(str, Enum):
@@ -48,7 +49,9 @@ class ZCCXT:
         None  # unable to set ccxt.Exchange type because all methods in that base class raise NotImplemented exception
     )
 
-    def __init__(self, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False):
+    def __init__(
+        self, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False
+    ):
         if exchange.value not in [e.value for e in ExchangeEnum]:
             raise ECInvalidValue(f"Invalid exchange provided: {exchange}")
 
@@ -60,7 +63,9 @@ class ZCCXT:
         self.db_disabled = disable_db
 
     @classmethod
-    async def create(cls, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False):
+    async def create(
+        cls, api_key: str, api_secret: str, exchange: ExchangeEnum, is_demo=False, identifier=None, disable_db=False
+    ):
         self = cls(api_key, api_secret, exchange, is_demo, identifier, disable_db)
 
         params = {
@@ -85,7 +90,7 @@ class ZCCXT:
             params.update({"uid": self.identifier})
 
         self.client = getattr(ccxt, self.exchange.lower())(params)
-        
+
         if not self.db_disabled:
             conn = connections.get("default")
             _, res = await conn.execute_query(
@@ -103,7 +108,7 @@ class ZCCXT:
             else:
                 self.client.set_markets(json.loads(key["data"]))
         else:
-            markets = self.client.load_markets(reload=True)
+            markets = await self.client.load_markets(reload=True)
             await self.cache_data(markets)
 
         if self.is_demo:
@@ -142,6 +147,9 @@ class ZCCXT:
             # for HL we need to fetch balance twice, once for futures and another time for spot
             tasks.append(self.client.fetch_balance(params={"type": "spot"}))
             tasks.append(self.get_positions())
+        elif self.exchange == ExchangeEnum.BINANCE:
+            tasks.append(self.client.fetch_balance(params={"type": "future"}))
+            tasks.append(self.get_positions())
 
         responses = await asyncio.gather(*tasks)
         response_data = {
@@ -158,6 +166,10 @@ class ZCCXT:
             response_data["account_data"] = {}  # NOTE: not available on HL
             response_data["balance_spot"] = responses[1]
             response_data["positions"] = responses[2]
+        elif self.exchange == ExchangeEnum.BINANCE:
+            response_data["account_data"] = {}
+            response_data["balance_future"] = responses[1]
+            response_data["positions"] = responses[2]
 
         return self.handle_balance_response(response_data)
 
@@ -218,6 +230,7 @@ class ZCCXT:
         order_id: str,
         market: typing.Optional[MarketTypeEnum] = MarketTypeEnum.SPOT,
         order_type: typing.Optional[typing.Union["open", "closed"]] = None,
+        symbol: typing.Optional[str] = None,
     ):
         try:
             if self.exchange == ExchangeEnum.BYBIT:
@@ -235,7 +248,10 @@ class ZCCXT:
                     with suppress(ccxt_errors.OrderNotFound):
                         return await self.client.fetch_open_order(order_id, params={"category": market_value})
                     return await self.client.fetch_closed_order(order_id, params={"category": market_value})
-            return await self.client.fetch_order(order_id)
+            elif self.exchange == ExchangeEnum.BINANCE:
+                return await self.client.fetch_order(order_id, symbol)
+            else:
+                return await self.client.fetch_order(order_id)
         except ccxt_errors.OrderNotFound:
             return None
         except ccxt_errors.BadRequest as e:
@@ -251,6 +267,7 @@ class ZCCXT:
         order_id: str,
         market: typing.Optional[MarketTypeEnum] = MarketTypeEnum.SPOT,
         order_type: typing.Optional[typing.Union["open", "closed"]] = None,
+        symbol: typing.Optional[str] = None,
     ):
         """
         Fetch an order from the exchange
@@ -303,11 +320,52 @@ class ZCCXT:
                 }
                 order["fees"] = fees_by_currency
 
+            return order
+        elif self.exchange == ExchangeEnum.BINANCE:
+            order = await self._get_order(order_id, symbol=symbol)
+            # set default values in case we return early
+            order.update(
+                {
+                    "average_price": 0,
+                    "fee": {},
+                    "fees": [],
+                }
+            )
+            # NOTE: sux that we need this, but Binance doesnt return fee info directly when fetching an order ...
+            if order["filled"] > 0:
+                trades = await self.get_order_trades(order_id, market, symbol=symbol)
+                if len(trades) == 0:
+                    log.warning("Didnt fetch any trades for HL order id %s", order_id)
+                    return order
+
+                average_price = sum(trade["price"] for trade in trades) / len(trades)
+                total_fee_cost = sum(trade["fee"]["cost"] for trade in trades)
+
+                # Sum of costs in 'fees', grouped by currency
+                fees_by_currency_dict = {}
+                for trade in trades:
+                    for fee in trade["fees"]:
+                        currency = fee["currency"]
+                        if currency not in fees_by_currency_dict:
+                            fees_by_currency_dict[currency] = {"currency": currency, "cost": 0}
+                        fees_by_currency_dict[currency]["cost"] += fee["cost"]
+
+                fees_by_currency = list(fees_by_currency_dict.values())
+
+                order["average"] = average_price
+                order["fee"] = {
+                    "currency": trades[0]["fee"]["currency"],
+                    "cost": total_fee_cost,
+                }
+                order["fees"] = fees_by_currency
+
             return order
         else:
             raise ECInvalidValue(f"Unknown exchange {self.exchange} provided")
 
-    async def get_order_trades(self, order_id, market: typing.Optional[MarketTypeEnum] = MarketTypeEnum.SPOT):
+    async def get_order_trades(
+        self, order_id, market: typing.Optional[MarketTypeEnum] = MarketTypeEnum.SPOT, symbol=None
+    ):
         """
         Fetch trades for an order
         """
@@ -337,6 +395,19 @@ class ZCCXT:
                     trade.update({"takerOrMaker": "maker"})
                 filtered_trades.append(trade)
             return filtered_trades
+        elif self.exchange == ExchangeEnum.BINANCE:
+            order = await self.client.fetch_order(order_id, symbol)
+            # timestamp attr returned directly by CCXT does uses a different value: statusTimestamp instead of timestamp
+            timestamp = int(order["timestamp"])
+
+            trades = await self.client.fetch_my_trades(order["symbol"], since=timestamp)
+            filtered_trades = []
+            for trade in trades:
+                if trade["order"] != order_id:
+                    continue
+
+                filtered_trades.append(trade)
+            return filtered_trades
         else:
             raise ECInvalidValue(f"Unknown exchange {self.exchange} provided")
 
@@ -632,6 +703,35 @@ class ZCCXT:
             # NOTE: we have no info about max page size from HL docs
             next_page = None
 
+        elif self.exchange == ExchangeEnum.BINANCE:
+            params = {}
+            results = await self.client.fetch_funding_history(symbol, from_dt, page_size, params=params)
+            for item in results:
+                timestamp = item["timestamp"]
+                rate_data = await self.client.fetch_funding_rate_history(
+                    symbol, since=timestamp, limit=1, params={"until": timestamp}
+                )
+                if len(rate_data) > 1:
+                    log.warning(
+                        "Got %s results when fetcing funding fees for % at timestamp %s",
+                        len(rate_data),
+                        symbol,
+                        timestamp,
+                    )
+                items.append(
+                    {
+                        "id": item["id"],
+                        "exchange": ExchangeEnum.BINANCE,
+                        "ticker_exchange": item["info"]["symbol"],
+                        "ticker_name": item["symbol"],
+                        "funding_rate": Decimal(rate_data[0]["fundingRate"]),
+                        "funding_fee": Decimal(item["amount"]),
+                        "created_at": datetime.utcfromtimestamp(float(timestamp) / 1000),
+                        "updated_at": datetime.utcfromtimestamp(float(timestamp) / 1000),
+                    }
+                )
+            # there is no pagination on binance, we have to implement our own
+            next_page = None
         else:
             raise ECInvalidValue(f"Unknown exchange {self.exchange} provided")
 
@@ -868,6 +968,63 @@ class ZCCXT:
                     "free": Decimal(str(response_spot["free"].get(token, 0) or 0)).quantize(Decimal("0.00000001")),
                 }
 
+        elif self.exchange == ExchangeEnum.BINANCE:
+            account_data = response_data["account_data"] = {}
+            positions = response_data["positions"]
+
+            balance_future = response_data["balance_future"]
+
+            total_notional = Decimal(sum(Decimal(position["notional"]) for position in positions)).quantize(
+                Decimal("0.00000001")
+            )
+            free_margin = Decimal(balance_future["info"]["availableBalance"])
+            total_balance = Decimal(balance_future["info"]["totalWalletBalance"])
+            margin_balance = Decimal(balance_future["info"]["totalMarginBalance"])
+            initial_margin = Decimal(balance_future["info"]["totalPositionInitialMargin"])
+            maintenance_margin = Decimal(balance_future["info"]["totalMaintMargin"])
+            maintenance_margin_ratio = maintenance_margin / total_balance if maintenance_margin else Decimal(0)
+            margin_ratio = (margin_balance / total_notional if total_notional else Decimal(0)).quantize(
+                Decimal("0.00000001")
+            )
+
+            balance.update(
+                {
+                    "account_type": None,  # account_data["accountMode"],
+                    "position_mode": "ONE_WAY",  # only ONE_WAY is possible on Binance??
+                    "total_balance": total_balance,
+                    "margin_balance": margin_balance,
+                    "free_margin": free_margin,
+                    "initial_margin": initial_margin,
+                    "total_notional": total_notional,
+                    "margin_ratio": margin_ratio,  # unified with bybit (account MR) as WOO returns a general MR under marginRatio
+                    "maintenance_margin": balance_future["info"]["totalMaintMargin"],
+                    "maintenance_margin_ratio": maintenance_margin_ratio,  # unified with bybit (account MMR) as WOO returns a general MMR under maintenanceMarginRatio
+                }
+            )
+
+            # spot balances
+            for token, val in response_data["balance"].items():
+                if not isinstance(val, dict):
+                    continue
+                if "free" not in val and "total" not in val:
+                    continue
+
+                total = Decimal(str(val["total"] or 0)).quantize(Decimal("0.00000001"))
+                free = Decimal(str(val["free"] or 0)).quantize(Decimal("0.00000001"))
+
+                if total <= Decimal(0) or free <= Decimal(0):
+                    continue
+
+                balance["coins"][token] = {
+                    "coin": token,
+                    "usdValue": None,
+                    "equity": None,
+                    "balance": None,
+                    # we have this to make it compatible with other dicts
+                    "total": total,
+                    "free": free,
+                }
+
         else:
             raise ECInvalidValue(f"Unknown exchange {self.exchange} provided")
 
@@ -889,6 +1046,9 @@ class ZCCXT:
         elif self.exchange == ExchangeEnum.WOO:
             spot_balance = await self.client.fetch_balance()
             futures_balance = spot_balance
+        elif self.exchange == ExchangeEnum.BINANCE:
+            spot_balance = await self.client.fetch_balance(params={"type": "spot"})
+            futures_balance = await self.client.fetch_balance(params={"type": "future"})
         else:
             raise ECInvalidValue(f"{self.exchange} not supported")
 
@@ -976,15 +1136,17 @@ class ZCCXT:
 
     def round_qty(self, symbol: str, qty):
         try:
-            proposed_qty = float(abs(qty))
+            proposed_qty = float(qty)
             qty = float(self.client.amount_to_precision(symbol, proposed_qty))
             # amount_to_precision does not round numbers down but it rounds it up, causing issue when selling SPOT
             # because we want to sell more than we have...
             # TIL: float(0.9) > Decimal("0.9") = True ... this is because float(0.9)  # ≈ 0.900000000000000022
-            if Decimal(str(qty)) > proposed_qty:
+            if Decimal(str(abs(qty))) > abs(proposed_qty):
                 amount_precision = Decimal(str(self.client.markets[symbol]["precision"]["amount"]))
-                # if amount_precision > 0:
-                #     amount_precision = int(amount_precision)
+                if amount_precision >= 1:
+                    # we need this for when we get "1.0" returned (eg for PURR/USDC), and if there's a trailing 0 in
+                    # the quantize will assume we want to have up to 1 decimal places...
+                    amount_precision = int(amount_precision)
                 if self.client.precisionMode == DECIMAL_PLACES:
                     amount_precision = Decimal("10") ** -amount_precision
                 qty = Decimal(proposed_qty).quantize(amount_precision, rounding=ROUND_DOWN)
@@ -994,4 +1156,7 @@ class ZCCXT:
         return qty
 
     def round_price(self, symbol, price):
-        return self.client.price_to_precision(symbol, float(price))
+        try:
+            return self.client.price_to_precision(symbol, float(price))
+        except ccxt_errors.InvalidOrder:
+            return 0
